unit iduoConsts;

interface

uses Windows, Classes, SysUtils, iduoResource, WordTokenizer, iduoToken;


const

// 䖽߂̃^O
CTRL_TAG      = #1;

// 䖽ߎʎq
CTRL_ID_BREAK = 'break';        // ubNo
CTRL_ID_EXIT  = 'exit';         // vZXI



type

EFunctionCallError = Exception;
EFunctionExecuteError = Exception;


TFunctionArgument = array of TVar;
TFunctionArgumentNames = array of string;
TFunctionProc = procedure(Args: TFunctionArgument; var return: TVar);

TFunctionData = record
  ArgCount: integer;
  Name: string;
  FuncType: (ftProc, ftScript);
  Proc: TFunctionProc;
  Script: TWordTokenArray;
  ArgNames: TFunctionArgumentNames;
end;


TFunctions = class
private
  FCount: integer;
  FSubFuncs: TList;
  Data: array of TFunctionData;
  function GetItem(index: integer): TFunctionData;
public
  constructor Create;
  destructor Destroy; override;
  procedure EntryFunction(Name: string; ArgCount: integer; Proc: TFunctionProc); overload;
  procedure EntryFunction(Name: string; ArgCount: integer; Proc: TWordTokenArray; ArgNames: TFunctionArgumentNames); overload;
  procedure AddSubFunction(funcs: TFunctions);
  procedure DeleteSubFunction(funcs: TFunctions);
  procedure ClearEntries;
  procedure Call(Name: string; Args: TFunctionArgument; var return: TVar);
  function IndexOf(Name: string): integer;
  function GetFunction(Name: string): TFunctionData;
  property Count: integer read FCount;
  property SubFuncs: TList read FSubFuncs;
  property Items[index: integer]: TFunctionData read GetItem; default;
end;


const
OT_Normal = 0;  // ̑̉Zq
OT_Subst  = 1;  // Zq
OT_Inc    = 2;  // C(f)NgZq

AT_BOTH_LR  = 0;         // E̒lAEɉZ
AT_BOTH_RL  = 1;         // E̒lAE獶
AT_RIGHT_RL = 2;         // E̒lAE獶
AT_LEFT_LR  = 3;         // ̒lAE

OPERATOR_LEVEL_MAX = 5;          // Zq̗D揇ʂ̍ől(Œxl)

type

TOperationProc = procedure(aLeft, aRight: TVar; var aReturn: TVar);

TOperationData = record
  Name:    string;
  Proc:    TOperationProc;
  Argtype: integer;             // K
  Level:   byte;                // D惌x(0:ŗD`OPERATOR_LEVEL_MAX)
  otype:   integer;             // Zq̃^Cv
end;

TOperators = class
private
  FCount: integer;
public
    Data: array of TOperationData;
  constructor Create;
  destructor Destroy; override;
  procedure EntryOperator(Name: string; Proc: TOperationProc; Argtype: integer; Level: integer; otype: integer=OT_Normal);
  procedure Call(Op: string; aLeft, aRight: TVar; var aReturn: TVar);
  function IndexOf(Name: string): integer;
  property Count: integer read FCount;
end;



var

{ 邵 }
MARK_FUNC: char = '`';
MARK_VAR: char = '$';
MARK_ARRAY: string[2] = '[]';
MARK_VAR_BRACKET: string[2] = '{}';


{ ʎq }
Operators: TOperators;        // Zq
Functions: TFunctions;        // ֐


implementation

uses iduoCode;


constructor TFunctions.Create;
begin
  inherited Create;
  FCount:=0;
  SetLength(Data, 128);
  FSubFuncs:=TList.Create;
end;

destructor TFunctions.Destroy;
begin
  FSubFuncs.Free;
  inherited Destroy;
end;

procedure TFunctions.EntryFunction(Name: string; ArgCount: integer; Proc: TFunctionProc);
var
  v: TFunctionData;
begin
  v.Name        := LowerCase(Name);
  v.ArgCount    := ArgCount;
  v.FuncType    := ftProc;
  v.Proc        := Proc;
  Data[FCount]  := v;
  inc(FCount);
end;

procedure TFunctions.EntryFunction(Name: string; ArgCount: integer; Proc: TWordTokenArray; ArgNames: TFunctionArgumentNames);
var
  v: TFunctionData;
begin
  v.Name        := LowerCase(Name);
  v.ArgCount    := ArgCount;
  v.FuncType    := ftScript;
  v.Script      := Proc;
  v.ArgNames    := ArgNames;
  Data[FCount]  := v;
  inc(FCount);
end;

procedure TFunctions.ClearEntries;
begin
  FCount:=0;
end;

procedure TFunctions.Call(Name: string; Args: TFunctionArgument; var return: TVar);
var
  i: integer;
  f: TFunctionData;
  mem: TVarMemory;
  subfuncs: TFunctions;
begin
  f:=GetFunction(Name);
  if f.Name<>'' then
  begin
    { vV[W }
    if f.FuncType=ftProc then
    try
      // ֐s
      f.Proc(Args, return)
    except
    on e: Exception do
      raise EFunctionExecuteError.Create(e.Message);
    end
    { XNvg }
    else begin
      // Tu֐
      subfuncs:=TFunctions.Create;
      RegisterFunctions(f.Script, subfuncs);
      Functions.AddSubFunction(subfuncs);

      // ̃f[^ϐɓ
      mem:=TVarMemory.Create;
      for i:=0 to f.ArgCount-1 do
        SetVar(mem.GetMem(f.ArgNames[i])^, Args[i]);
      // ֐s
      ExecuteProcedure(f.Script, mem, return);

      Functions.DeleteSubFunction(subfuncs);
      mem.Free;
      subfuncs.Free;
    end;
  end else
    raise EFunctionCallError.Create('֐݂܂');
end;

function TFunctions.IndexOf(Name: string): integer;
var
  i: integer;
begin
  Name := LowerCase(Name);
  for i:=0 to FCount-1 do
  if Data[i].Name=Name then
  begin
    Result:=i;
    Exit;
  end;
  Result:=-1;
end;

function TFunctions.GetFunction(Name: string): TFunctionData;
var
  i: integer;
  r: TFunctionData; 
begin
  for i:=0 to SubFuncs.Count-1 do
  begin
    r:=TFunctions(SubFuncs[i]).GetFunction(Name);
    if r.Name<>'' then
    begin
      Result:=r;
      Exit;
    end;
  end;

  i:=IndexOf(Name);
  if i>=0 then
  begin
    Result:=Data[i];
    Exit;
  end;

  r.ArgCount:=0;
  r.Name:='';
  Result:=r;
end;

procedure TFunctions.AddSubFunction(funcs: TFunctions);
begin
  SubFuncs.Add(funcs);
end;

procedure TFunctions.DeleteSubFunction(funcs: TFunctions);
var i: integer;
begin
  i:=SubFuncs.IndexOf(funcs);
  if i>=0 then SubFuncs.Delete(i);
end;

function TFunctions.GetItem(index: integer): TFunctionData;
begin
  Result:=Data[index];
end;









constructor TOperators.Create;
begin
  inherited Create;
  FCount:=0;
  SetLength(Data, 128);
end;

destructor TOperators.Destroy;
begin
  inherited Destroy;
end;

procedure TOperators.EntryOperator(Name: string; Proc: TOperationProc; Argtype: integer; Level: integer; otype: integer=OT_Normal);
var
  v: TOperationData;
begin
  v.Name        := LowerCase(Name);
  v.Argtype     := Argtype;
  v.Proc        := Proc;
  v.Level       := Level;
  v.otype       := otype;
  Data[FCount]  := v;
  inc(FCount);
  RegisterOperator(Name);
end;

procedure TOperators.Call(Op: string; aLeft, aRight: TVar; var aReturn: TVar);
var
  i: integer;
begin
  i:=IndexOf(Op);
  if i>=0 then
  begin
    Data[i].Proc(aLeft, aRight, aReturn);
  end else
    raise EFunctionCallError.Create('Zsł܂');
end;

function TOperators.IndexOf(Name: string): integer;
var
  i: integer;
begin
  Name := LowerCase(Name);
  for i:=0 to FCount-1 do
  if Data[i].Name=Name then
  begin
    Result:=i;
    Exit;
  end;
  Result:=-1;
end;




initialization

Operators := TOperators.Create;
Functions := TFunctions.Create;

finalization

Operators.free;
Functions.free;


end.
