unit iduoToken;

interface

uses Windows, Classes, SysUtils, Variants, WordTokenizer, iduoResource;



const

// Zg[N^Cv
PT_Invalid        = -1;
PT_Variable       = 0;
PT_String         = 1;
PT_Numerical      = 2;
PT_Value          = 3;      // oAg̒l
PT_Operator       = 4;
PT_Function       = 5;
PT_Nest           = 6;

type

TProcToken = record
  word: TWordTokenArray;
  ttype: integer;
  value: TVar;               // ttype=PT_Valuêݐݒ
end;
PProcToken = ^TProcToken;

TProcTokenList = class
private
  List: TList;
  function GetItem(index: integer): PProcToken;
  function GetCount: integer;
public
  constructor Create;
  destructor Destroy; override;
  procedure Clear;
  function Add(const w: TWordTokenArray; TokenType: integer; value: TVar): PProcToken; overload;
  function Add(const t: TProcToken): PProcToken; overload;
  function IndexOf(t: PProcToken): integer;
  procedure Delete(Index: integer);
  property Items[Index: integer]: PProcToken read GetItem; default;
  property Count: integer read GetCount;
end;


// g[NG[
ETokenError = class(Exception)
public
    word: string;
    pos: integer;
    line: integer;
  constructor Create(msg: string; word: string; pos, line: integer);
end;


// g[ÑlXg
function NestProcToken(const t: TWordTokenArray; index: integer; bracket1: widestring='('; bracket2: widestring=')'): integer;
// g[N
function ProcTokenize(const t: TWordTokenArray): TProcTokenList;
// g[ÑRs[
procedure CopyProcToken(var dest: TProcToken; const src: TProcToken);
// g[N̊J
procedure FreeToken(var dest: TProcToken);


implementation

uses iduoParser, iduoConsts;



constructor ETokenError.Create(msg: string; word: string; pos, line: integer);
begin
    self.word := word;
    self.pos  := pos;
    self.line := line;
  inherited Create(msg);
end;



constructor TProcTokenList.Create;
begin
  inherited Create;
    List:=TList.Create;
    List.Capacity:=16;
end;

destructor TProcTokenList.Destroy;
begin
  Clear;
  List.free;
  inherited Destroy;
end;

procedure TProcTokenList.Clear;
begin
  while Count>0 do Delete(0);
end;

function TProcTokenList.Add(const w: TWordTokenArray; TokenType: integer; value: TVar): PProcToken;
var r: PProcToken;
begin
  new(r);
  r.word  := w;
  r.ttype := TokenType;
  r.value := value;
  List.Add(r);
  Result:=r;
end;

function TProcTokenList.Add(const t: TProcToken): PProcToken;
var r: PProcToken;
begin
  new(r);
  r^:=t;
  r^.value:=MakeNullVar;
  SetVar(r^.value, t.value);
  List.Add(r);
  Result:=r;
end;

procedure TProcTokenList.Delete(Index: integer);
begin
  FreeVar(Items[index].value);
  dispose(List[Index]);
  List.Delete(Index);
end;

function TProcTokenList.GetItem(index: integer): PProcToken;
begin
  Result:=List[index];
end;

function TProcTokenList.IndexOf(t: PProcToken): integer;
begin
  Result:=List.IndexOf(t);
end;

function TProcTokenList.GetCount: integer;
begin
  Result:=List.Count;
end;



{ g[ÑlXg }
// ߂l: w肳ꂽ͂(uPbg)̏I܂ł̃g[N
function NestProcToken(const t: TWordTokenArray; index: integer; bracket1: widestring='('; bracket2: widestring=')'): integer;
   var
     i, c: integer;
     ret: string;
     fs: integer;
begin
     c:=Length(t);
     i:=index;
     ret:='';
        // ̏I܂Ńg[N
        if (i<c)and(t[i].Str=bracket1) then
        begin

          fs:=0;
          repeat
            if i<c then
            begin
              if t[i].Str=bracket1 then
                inc(fs)
              else if t[i].Str=bracket2 then
                dec(fs);
            end else
              raise ETokenError.Create('"'+bracket2+'"Kvł', t[index].Str, t[index].Pos, t[index].Line);

            ret:=ret+t[i].Str;
            inc(i);
          until fs=0;

        end else
          raise ETokenError.Create('sȕ', t[i].Str, t[i].Pos, t[index].Line);

     Result:=i-index;
end;


{ g[Nɕ }

//1. $b=replace($a,$b+$c,"\$1")
//2. $b/=/`replace($a,$b+$c,"\$1")

function ProcTokenize(const t: TWordTokenArray): TProcTokenList;
var
 i, c: integer;
 ret: TProcTokenList;   // g[N
 tNum: integer;         // g[N̐

   procedure AddToken(const w: TWordTokenArray; ttype: integer; value: TVar);
   begin
        inc(tNum);
        ret.Add(w, ttype, value);
   end;

var
  j: integer;
begin
  c:=Length(t);
  ret:=TProcTokenList.Create;
  tNum:=0;
  i:=0;

  while (i<c) do
  begin

    // ϐ
    if t[i].Str=MARK_VAR then
    begin

        if (i+1<c)and(t[i+1].Str=MARK_VAR_BRACKET[1]) then
        begin
          j:=NestProcToken(t, i+1, MARK_VAR_BRACKET[1], MARK_VAR_BRACKET[2]);
          AddToken(SubToken(t, i, j+1), PT_Variable, MakeNullVar());
          inc(i, j+1);
        end else        
        if (i+1<c)and(t[i+1].WordType=WT_Identifier)and(t[i].Pos = t[i+1].Pos-1) then
        begin
          j:=1;
          // ** 06/05/03 ύX
          // z
          while (i+j+3<c)and(t[i+j+1].Str=MARK_ARRAY[1]) do
          begin
            j:=j+NestProcToken(t, i+j+1, MARK_ARRAY[1], MARK_ARRAY[2]);
          end;

          AddToken(SubToken(t, i, j+1), PT_Variable, MakeNullVar());
          inc(i, j+1);
          
        end else
          raise ETokenError.Create('sȕ', t[i].Str, t[i].Pos, t[i].Line);
    end
    // 
    else if t[i].WordType=WT_STRING then
    begin
      AddToken(SubToken(t, i, 1), PT_Value, MakeVar('', Copy(t[i].Str, 2, Length(t[i].Str)-2)));
      inc(i);
    end
    // 
    else if t[i].WordType=WT_Numerical then
    begin
      AddToken(SubToken(t, i, 1), PT_Value, MakeVar('', StrToFloat(t[i].Str)));
      inc(i);
    end
    // u[A萔
    else if (t[i].WordType=WT_Identifier) and ((LowerCase(t[i].Str)='true') or (LowerCase(t[i].Str)='false')) then
    begin
      AddToken(SubToken(t, i, 1), PT_Value, MakeVar('', (LowerCase(t[i].Str)='true')));
      inc(i);
    end
    // Zq
    else if Operators.IndexOf(LowerCase(t[i].Str))>=0 then
    begin
      AddToken(SubToken(t, i, 1), PT_Operator, MakeNullVar());
      inc(i);
    end
    // ֐
    else if Functions.GetFunction(LowerCase(t[i].Str)).Name<>'' then
    begin
      j:=i;
        // ̃lXg
        if (i+1<c)and(t[i+1].Str='(') then
        begin
          inc(i);
          i:=i+NestProcToken(t, i);
        end else
          inc(i);
          
        AddToken(SubToken(t, j, i-j), PT_Function, MakeNullVar());
    end
    // 
    else if t[i].Str ='(' then
    begin

      j:=NestProcToken(t, i);
      AddToken(SubToken(t, i, j), PT_Nest, MakeNullVar());
      inc(i, j);

    end else
    if not (t[i].WordType=WT_Space) then
      raise ETokenError.Create('`̎ʎq', t[i].Str, t[i].Pos, t[i].Line)
    else
      inc(i);

  end;

  // CNgZq
  c:=ret.Count;
  i:=c-2;
  while i>=0 do
  begin
    if (ret[i].ttype=PT_Operator)and(ret[i+1].ttype=PT_Variable) then
    begin
      j:=Operators.IndexOf(ret[i].word[0].Str);
      if Operators.Data[j].otype=OT_Inc then
      begin
        ret[i].word:=ConcatToken(ret[i].word, ret[i+1].word);
        ret[i].ttype:=PT_Variable;
        ret.Delete(i+1);
      end;
    end

    else if (ret[i].ttype=PT_Variable)and(ret[i+1].ttype=PT_Operator) then
    begin
      j:=Operators.IndexOf(ret[i+1].word[0].Str);
      if Operators.Data[j].otype=OT_Inc then
      begin
        ret[i].word:=ConcatToken(ret[i].word, ret[i+1].word);
        ret.Delete(i+1);
      end;
    end;

    dec(i);
  end;

  tNum:=ret.Count;

  result:=ret;
end;


procedure CopyProcToken(var dest: TProcToken; const src: TProcToken);
begin
  dest.word := src.word;
  dest.ttype := src.ttype;
  SetVar(dest.value, src.value);
end;


procedure FreeToken(var dest: TProcToken);
begin
  FreeVar(dest.value);
  ZeroMemory(@dest, sizeof(TProcToken));
end;



end.
